#include <string.h>

#include "_version.h"

#include "fk_help.h"

static const char versiontext[] =
    "fedakeys program: key management for FEDAnet users and node admins\n"
    "vers. " FEDA_VERSION " (compiled " __DATE__ ")\n"
    "Copyright (c) Andrey Vikt. Stolyarov, 2024\n"
    ;

static const char helptext[] =
    "\n"
    "Usage: fedakeys <global-options> command <command-args>\n"
    "Options are:\n"
    "    -c <directory>      use this dir instead of $HOME/.fedanet/keys\n"
    "    -o <output_file>    override the default output file name\n"
    "    -t <timemark>       use this as the timemark instead of the\n"
    "                        current time; remember it's in MINUTES since\n"
    "                        the Epoch, that is, unixtime/60; '0x' prefix\n"
    "                        means hex, leading '0' means octal\n"
    "    -v                  be verbose (more flags for more messages)\n"
    "    -q                  be quiet\n"
    "    -h                  print help text and quit; if a command is\n"
    "                        given, the help on that command is printed;\n"
    "                        other options are ignored in this case.\n"
    "\n"
    "Supported commands are briefly explained below; please note that\n"
    "command names starting with ``m'' are intended for a node master\n"
    "and are only useful in presence of your node master secret key,\n"
    "which is not (and MUST NOT be) a very common situation.  Keep your\n"
    "node master secret key away from computers connected to networks!\n"
    "Use ``fedakeys -h <command>'' for command-specific help\n"
    "\n"
    "    mverify             verify a file generated by feda-ng\n"
    "    mdeploy             deploy nodemaster workplace\n"
    "    mzero               make zero (0th) point key\n"
    "    mpoint              make a key for the given point number\n"
    "\n"
    "    pverify             verify a point file\n"
    "    deploy              deploy point\n"
    "    zcrpoint            make a key for another point using Zero Point\n"
    "    zsign               sign an existing public key as belonging to\n"
    "                        the given point using Zero Point\n"
    "    keypair             make a pair of secret and public keys\n"
    "    kexgen              (re)generate the key exchange keypair\n"
    "    export              compose a file with public keys and certs\n"
    "                        for the locally configured point\n"
    "    sign                sign a given file with the point's signature\n"
    "    signverify          check the signature for the file\n"
    "    lock                enclose the given file in a ciphered container\n"
    "    anonlock            use a one-time keypair to enclose the given\n"
    "                        file in a ciphered container\n"
    "    unlock              check and open a container\n"
    ;

static const char mverify_helptext[] =
    "fedakeys <options> mverify <master_key_file>\n"
    "    Verify a file generated by the feda-ng program for integrity.\n"
    ;

static const char mdeploy_helptext[] =
    "fedakeys <options> mdeploy <master_key_file> [<min_pref_len>]\n"
    "    Deploy nodemaster workplace using the file generated by the\n"
    "    feda-ng program, setting the given minimal prefix length for\n"
    "    foreign node keys to be trusted.\n"
    "    If the min_pref_len arg is omitted, the prefix length of the\n"
    "    key being deployed is used as the minimum (may be changed later)\n"
    ;

static const char mzero_helptext[] =
    "fedakeys <options> mzero\n"
    "    Generate a key pair and create a certificate for the current\n"
    "    node's Point Zero (0th point), which may be used to sign other\n"
    "    points' certificates and other things on behalf of the node.\n"
    "    Requires deployed nodemaster workplace.\n"
    "    All the data is saved in the file named NODEID_p0.key, where\n"
    "    NODEID is the 20-hex-digits node ID.\n"
    ;

static const char mpoint_helptext[] =
    "fedakeys <options> mpoint <point>\n"
    "    Generate a key pair and create a certificate for the given\n"
    "    point using the master key to sign it.  <point> must be a decimal\n"
    "    number from 1 to 253.  Deployed nodemaster workplace is required.\n"
    "    All the data is saved in the file named NODEID_pNUM.key, where\n"
    "    NODEID is the 20-hex-digits node ID and NUM is the point number.\n"
    "    Generally this is discouraged practice; Zero Point workplace\n" 
    "    and keys should be used to sign all point certificates, so the\n"
    "    master key is accessed less often.  However, replacing your Zero\n"
    "    Point key with a newer one invalidates all certs signed with\n"
    "    the old key, so sometimes it might make sense to create some keys\n"
    "    for your points using the master key while you're at it anyway.\n"
    ;

static const char pverify_helptext[] =
    "fedakeys <options> pverify <point_key_file>\n"
    "    Verify a key file created by ``mzero'', or ``mpoint'',\n"
    "    or ``zcrpoint'', or ``zsign'' command.\n"
    ;

static const char deploy_helptext[] =
    "fedakeys <options> deploy <point_key_file> [<separate_secret_file>]\n"
    "    Deploy a point's key and configuration using a key file created\n"
    "    by either ``mzero'', or ``mpoint'', or ``zcrpoint'', or ``zsign''\n"
    "    command.\n"
    ;

static const char zcrpoint_helptext[] =
    "fedakeys <options> zcrpoint <point>\n"
    "    Generate a key pair and create a certificate for the given\n"
    "    point using the Zero Point key to sign it.  <point> must be a\n"
    "    decimal number from 1 to 253.  Deployed Zero Point is required.\n"
    "    All the data is saved in the file named NODEID_pNUM.key, where\n"
    "    NODEID is the 20-hex-digits node ID and NUM is the point number.\n"
    ;

static const char zsign_helptext[] =
    "fedakeys <options> zsign <file> <point>\n"
    "    Take the given public key and sign it with the Zero Point key;\n"
    "    append all the necessary fields to the given file so it becomes\n"
    "    suitable for point deployment.\n"
    "    The file must initially contain the \"public\" field and SHOULD NOT\n"
    "    contain anything else.  The command modifies the file so you might\n"
    "    want to make a copy of it, just in case.\n"
    ;

static const char keypair_helptext[] =
    "fedakeys <options> keypair <basename>\n"
    "    Generate a keypair (suitable, e.g., for the zsign command).\n"
    "    Two files, <basename>.key and <basename>.pub, will be created.\n"
    "    No error checking is performed, so be careful not to write over\n"
    "    any files you still need.\n"
    "    In case the output file name is specified with -o, the <basename>\n"
    "    is ignored (and may be omitted), and both secret and public keys\n"
    "    are written to that file.\n"
    ;

static const char kexgen_helptext[] =
    "fedakeys <options> kexgen\n"
    "    (Re)generate the keypair intended for key exchange.\n"
    "    The key pair is always saved in the $dir/kex.key file, where $dir\n"
    "    is either specified with -c, or (by default) is the\n"
    "    $HOME/.fedanet/keys directory.\n"
    "    If the file already exists, it is renamed to kex_IDIDIDIDID.key,\n"
    "    where IDIDIDIDID is the first 10 hex digits of the public key.\n"
    ;

static const char export_helptext[] =
    "fedakeys <options> export\n"
    "    Compose a file with public keys and appropriate certs for the\n"
    "    locally deployed point.  The file is appropriate to be published\n"
    "    openly, and/or sent directly to people you're going to communicate\n"
    "    with.  If there's no -o option, the file is named NODEID_pNUM.pub,\n"
    "    where NODEID is your node ID and NUM is your point number.\n"
    "    If you don't yet have the keypair for message encryption (as\n"
    "    distinct from signatures), the pair will be generated and saved.\n"
    ;

static const char sign_helptext[] =
    "fedakeys <options> sign\n"
    "    NOT IMPLEMENTED YET\n"
    ;

static const char signverify_helptext[] =
    "fedakeys <options> signverify\n"
    "    NOT IMPLEMENTED YET\n"
    ;

static const char lock_helptext[] =
    "fedakeys <options> lock\n"
    "    NOT IMPLEMENTED YET\n"
    ;

static const char anonlock_helptext[] =
    "fedakeys <options> anonlock\n"
    "    NOT IMPLEMENTED YET\n"
    ;

static const char unlock_helptext[] =
    "fedakeys <options> unlock\n"
    "    NOT IMPLEMENTED YET\n"
    ;


#define HELPFOR(name)                     \
    if(0 == strcmp(cmd, #name)) {         \
        fputs(name##_helptext, stream);   \
        return;                           \
    }

void help(FILE *stream, const char *cmd)
{
    if(!cmd) {
        fputs(versiontext, stream);
        fputs(helptext, stream);
        return;
    }
    if(!*cmd) {
        fputs(helptext, stream);
        return;
    }
    HELPFOR(mverify)
    HELPFOR(mdeploy)
    HELPFOR(mzero)
    HELPFOR(mpoint)
    HELPFOR(pverify)
    HELPFOR(deploy)
    HELPFOR(zcrpoint)
    HELPFOR(zsign)
    HELPFOR(keypair)
    HELPFOR(kexgen)
    HELPFOR(export)
    HELPFOR(sign)
    HELPFOR(signverify)
    HELPFOR(lock)
    HELPFOR(anonlock)
    HELPFOR(unlock)
    fprintf(stream, "Command %s unknown, try -h\n", cmd);
}
