#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <monocypher/monocypher.h>

#include "hexdata.h"
#include "message.h"
#include "txtr_out.h"
#include "getrand.h"
#include "fk_args.h"
#include "fk_data.h"
#include "fk_mesg.h"


static int get_config(const char *dir, struct config_file *conf)
{
    char *fname;
    int res;

    fname = make_config_fname(dir);
    res = read_config_file(fname, conf);
    dispose_fname(fname);
    if(!res) {
        /* error message already printed by the function */
        message(mlv_info, "do you have the workplace deployed here?\n");
        return 0;
    }
    if(conf->point == -1) {
        message(mlv_alert, "can't do this at the master workplace\n");
        return 0;
    }
    return 1;
}

int main_kexgen(const struct fedakeys_args *args)
{
    int res;
    char *main_kex_fn;
    struct kex_keypair_file kex_data;
    FILE *f;
    int saved_umask;
    unsigned char kex_secret[kex_secret_size];
    unsigned char kex_public[kex_public_size];

    main_kex_fn = make_msgkey_fname(args->dir, NULL);
    res = read_kex_keypair_file(main_kex_fn, &kex_data, 1);
    if(res) {   /* this means it's already there so we need to move it */
        char *targ_fn;
        char buf[20];
        hexdata2str(buf, kex_data.kex_public, (sizeof(buf)-1)/2);
        targ_fn = make_msgkey_fname(args->dir, buf);
        res = rename(main_kex_fn, targ_fn);
        if(res == -1) {
            message_perror(mlv_alert, "FATAL: rename", main_kex_fn);
            dispose_fname(targ_fn);
            res = 1;
            goto quit;
        }
        f = fopen(targ_fn, "a");
        if(!f) {
            message_perror(mlv_alert, "WARNING", targ_fn);
            message(mlv_alert, "couln't write ``replaced'' timestamp");
        } else {
            put_int_field(f, "replaced", args->timestamp);
            fclose(f);
        }
        dispose_fname(targ_fn);
    }

    /* the way is clear, we can work */

    res = get_random(kex_secret, sizeof(kex_secret));
    if(!res) {
        message(mlv_alert, "can't read random numbers\n");
        res = 1;
        goto quit;
    }

    crypto_x25519_public_key(kex_public, kex_secret);

    saved_umask = umask(0077);
    f = fopen(main_kex_fn, "w");
    umask(saved_umask);
    if(!f) {
        message_perror(mlv_alert, "FATAL", main_kex_fn);
        return 1;
    }
    put_hex_field(f, "kex_secret", kex_secret, sizeof(kex_secret));
    put_hex_field(f, "kex_public", kex_public, sizeof(kex_public));
    put_int_field(f, "created", args->timestamp);
    fclose(f);
    res = 0;

quit:
    dispose_fname(main_kex_fn);
    return res;
}

int main_export(const struct fedakeys_args *args)
{
    struct config_file conf;
    int res;
    char output_fname[32]; /* 20 for node id, 7..9 for '_pNNN.pub', 2 resvd */
    char *fname;
    const char *ofn;
    FILE *f;
    char *main_kex_fn;
    unsigned char secret_key[secret_key_size];
    unsigned char kex_signature[signature_size];
    struct kex_keypair_file kex_data;
    struct feda_kex_info kex_info;

    res = get_config(args->dir, &conf);
    if(!res)    /* everything has already been said by get_config */
        return 1;

    main_kex_fn = make_msgkey_fname(args->dir, NULL);
    res = read_kex_keypair_file(main_kex_fn, &kex_data, 0);
    dispose_fname(main_kex_fn);

    if(!res) {
        message(mlv_alert,
            "no key-exchg key found, use the ``kexgen'' command to create");
        return 1;
    }

    compose_kex_info(&kex_info, args->timestamp, kex_data.kex_public);

    fname = make_secretkey_fname(args->dir);
    res = read_single_blob(fname, "secret", secret_key, sizeof(secret_key));
    dispose_fname(fname);
    if(!res) {
        message(mlv_alert, "can't read the point's secret key\n");
        return 1;
    }

    crypto_eddsa_sign(kex_signature, secret_key,
                      (unsigned char *)&kex_info, sizeof(kex_info));
    crypto_wipe(secret_key, sizeof(secret_key));

    if(args->out_file) {
        ofn = args->out_file;
    } else {
        hexdata2str(output_fname, conf.node_id, sizeof(conf.node_id));
        sprintf(output_fname+2*node_id_size, "_p%d.pub", conf.point);
        ofn = output_fname;
    }

    f = fopen(ofn, "w");
    if(!f) {
        message_perror(mlv_alert, "FATAL", ofn);
        return 1;
    }

    put_hex_field(f, "node_id", conf.node_id, node_id_size);
    put_int_field(f, "point", conf.point);
    put_int_field(f, "minpref", conf.minpref);
    put_int_field(f, "signed_by", conf.signed_by);
    put_hex_field(f, "master_pub", conf.master_pub, public_key_size);
    put_hex_field(f, "master_hash", conf.master_hash, yespower_hash_size);
    if(conf.signed_by == 0) {
        put_hex_field(f, "zp_certbody", conf.zp_certbody,
                                 sizeof(conf.zp_certbody));
        put_hex_field(f, "zp_signature", conf.zp_signature,
                                 sizeof(conf.zp_signature));
    }
    put_hex_field(f, "certbody", conf.certbody, sizeof(conf.certbody));
    put_hex_field(f, "signature", conf.signature, sizeof(conf.signature));
    put_hex_field(f, "public", conf.public_key, sizeof(conf.public_key));
    put_hex_field(f, "kex_public", kex_data.kex_public,
                                   sizeof(kex_data.kex_public));
    put_hex_field(f, "kex_cert", (unsigned char *)&kex_info,
                                 sizeof(kex_info));
    put_hex_field(f, "kex_signature", kex_signature, sizeof(kex_signature));
    put_int_field(f, "kex_timestamp", args->timestamp);


    fclose(f);

    return 0;
}

int main_sign(const struct fedakeys_args *args)
{
    message(mlv_alert, "NOT IMPLEMENTED YET\n");
    return 1;
}

int main_signverify(const struct fedakeys_args *args)
{
    message(mlv_alert, "NOT IMPLEMENTED YET\n");
    return 1;
}

int main_lock(const struct fedakeys_args *args)
{
    message(mlv_alert, "NOT IMPLEMENTED YET\n");
    return 1;
}

int main_anonlock(const struct fedakeys_args *args)
{
    message(mlv_alert, "NOT IMPLEMENTED YET\n");
    return 1;
}

int main_unlock(const struct fedakeys_args *args)
{
    message(mlv_alert, "NOT IMPLEMENTED YET\n");
    return 1;
}

