#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <monocypher/monocypher.h>

#include "message.h"
#include "fileutil.h"
#include "getrand.h"
#include "txtr_out.h"
#include "hexdata.h"
#include "fk_args.h"
#include "fk_data.h"
#include "fk_crypt.h"
#include "fk_mastr.h"  /* for check_master_pub */
#include "fk_point.h"

static int read_verify_point(const char *fname, const char *secret_fname,
                             struct point_file *pf, const char **err)
{
    int res, preflen;
    
    res = read_point_file(fname, pf);
    if(!res) {
        *err = "the file is not in proper format";
        return 0;
    }

    if(all_zeroes(pf->secret, secret_key_size)) {
        if(!secret_fname) {
            *err = "no secret key in the file and no separate key file name";
            return 0;
        }
        res = read_single_blob(secret_fname, "secret",
                               pf->secret, secret_key_size);
        if(!res) {
            *err = "couldn't get the secret key from the separate file";
            return 0;
        }
    }

    res = check_keypair(pf->secret, pf->public_key);
    if(!res) {
        *err = "public and secret keys don't match";
        return 0;
    }

    preflen = check_master_pub(pf->node_id, pf->master_pub, pf->master_hash);
    switch(preflen) {
    case checkmasterpub_hash_fail:
        *err = 0;
        return 0;
    case checkmasterpub_wrong_id:
        *err = "declared node ID doesn't match the key";
        return 0;
    case checkmasterpub_wrong_hash:
        *err = "computed hash doesn't match";
        return 0;
    default:
        break;
    }

    if(preflen < pf->minpref) {
        *err = "the master's 0xFEDA prefix is shorter than the minimum";
        return 0;
    }

    res = compare_feda_cert((struct feda_cert_info *)pf->certbody,
        pf->node_id, pf->point, pf->signed_by, pf->timestamp,
        pf->public_key);
    if(!res) {
        *err = "certbody malformed";
        return 0;
    }

    if(pf->signed_by == -1) { /* signed by master, this is simple */
        res = crypto_eddsa_check(pf->signature, pf->master_pub,
                                 pf->certbody, feda_cert_size);
        if(res /* sic! */) {
            *err = "signature check failed";
            return 0;
        }
    } else
    if(pf->signed_by == 0) { /* signed by Zero Point, check in two steps */
        res = crypto_eddsa_check(pf->zp_signature, pf->master_pub,
                                 pf->zp_certbody, feda_cert_size);
        if(res /* sic! */) {
            *err = "zero point certificate signature check failed";
            return 0;
        }
        res = compare_feda_cert((struct feda_cert_info *)pf->zp_certbody,
            pf->node_id, 0, -1, -1 /* ignore timestamp */, pf->zp_pub);
        if(!res) {
            *err = "zero point certbody malformed";
            return 0;
        }
        res = crypto_eddsa_check(pf->signature, pf->zp_pub,
                                 pf->certbody, feda_cert_size);
        if(res /* sic! */) {
            *err = "certificate signature check failed";
            return 0;
        }
    } else {
        *err = "unsupported signed_by value";
        return 0;
    }

    /* all checks passed, everything's fine */
    return 1;
}

int main_pverify(const struct fedakeys_args *args)
{
    struct point_file mf;
    int res;
    const char *err = "";  /* initialized to avoid a stupid warning */
    
    if(args->cmdargc < 2) {
        message(mlv_alert, "\"fedakeys mverify\" needs a file name, try -h\n");
        return 1;
    }
    res = read_verify_point(args->cmdargs[1], args->cmdargs[2], &mf, &err);
    if(!res) {
        message(mlv_normal, "VERIFICATION FAILED: %s\n", err);
        return 1;
    }

    message(mlv_normal, "verification passed\n");
    return 0;
}

static int
can_upgrade_zp(const struct config_file *conf, const struct point_file *pf)
{
    if(pf->point == 0 || conf->point != 0)
        return 0;
    if(0 != memcmp(conf->node_id, pf->node_id, node_id_size))
        return 0;
    if(pf->signed_by == 0) {
        /* it must be the same version of the Zero Point key */
        if(0 != memcmp(conf->public_key, pf->zp_pub, public_key_size))
            return 0;
    }
    return 1;
}

int main_deploy(const struct fedakeys_args *args)
{
    struct point_file mf;
    int res;
    const char *err;
    char *secretkey_fname, *config_fname;
    FILE *f;
    int saved_umask;
    int zeropoint_upgrade = 0;
    
    if(args->cmdargc < 2) {
        message(mlv_alert, "\"fedakeys deploy\" needs a file name, try -h\n");
        return 1;
    }

        /* now we grab all we need, and only quit via 'quit:' from now on */

    secretkey_fname = make_secretkey_fname(args->dir);
    config_fname = make_config_fname(args->dir);
    saved_umask = umask(0077);
    res = read_verify_point(args->cmdargs[1], args->cmdargs[2], &mf, &err);
    if(!res) {
        message(mlv_alert, "point file verification FAILED: %s\n", err);
        res = 1;
        goto quit;
    }

    if(file_is_there(config_fname)) {
        struct config_file conf;
        char *zeropointkey_fname;
        res = read_config_file(config_fname, &conf);
        if(!res) {
            message(mlv_alert, "FATAL: can't read %s, despite it exists\n",
                    config_fname);
            res = 1;
            goto quit;
        }
        if(!can_upgrade_zp(&conf, &mf)) {
            message(mlv_alert,
                    "FATAL: %s already exists, move/delete it and retry\n",
                    config_fname);
            res = 1;
            goto quit;
        }
        /* okay, we'll upgrade ZP-only configuration to P+ZP */
        zeropoint_upgrade = 1;
        zeropointkey_fname = make_zeropointkey_fname(args->dir);
        res = rename(secretkey_fname, zeropointkey_fname);
        dispose_fname(zeropointkey_fname);
        if(res == -1) {
            message_perror(mlv_alert, "FATAL: rename", secretkey_fname);
            res = 1;
            goto quit;
        }
        /* if we don't have the zero point info within the point file,
           which is the case if it is signed by the node's master key,
           we need to keep zero point info from the current config file
         */
        if(mf.signed_by != 0) {
            /* memcpy(mf.zp_pub, conf.public_key, public_key_size); */
            memcpy(mf.zp_certbody, conf.certbody, feda_cert_size);
            memcpy(mf.zp_signature, conf.signature, signature_size);
        }
    }

    if(file_is_there(secretkey_fname)) {
        message(mlv_alert,
                "FATAL: %s already exists, move/delete it and retry\n",
                secretkey_fname);
        res = 1;
        goto quit;
    }

    res = make_directory_path(args->dir, 0);
    if(res == -1) {
        message_perror(mlv_alert, "FATAL", args->dir);
        res = 1;
        goto quit;
    }

    f = fopen(secretkey_fname, "w");
    if(!f) {
        message_perror(mlv_alert, "FATAL", secretkey_fname);
        res = 1;
        goto quit;
    }
    put_hex_field(f, "secret", mf.secret, sizeof(mf.secret));
    fclose(f);

/* We might want to reimplement this using the write_config_file
   function, but the problem is that we'll have to copy all fields
   from the point file structure to the config file structure...
   which is weird.
       One possible reimplementation to be done one day is to
   get rid of all these different structures and have a single one
   with all possible fields, some of them remaining unused in
   different situations.
       Well... not today.
*/
    f = fopen(config_fname, "w");
    if(!f) {
        message_perror(mlv_alert, "FATAL", config_fname);
        res = 1;
        goto quit;
    }
    put_hex_field(f, "node_id", mf.node_id, sizeof(mf.node_id));
    put_int_field(f, "point", mf.point);
    put_int_field(f, "minpref", mf.minpref);
    put_hex_field(f, "public", mf.public_key, sizeof(mf.public_key));
    put_int_field(f, "signed_by", mf.signed_by);
    put_hex_field(f, "master_pub", mf.master_pub, sizeof(mf.master_pub));
    put_hex_field(f, "master_hash", mf.master_hash, sizeof(mf.master_hash));
/*  put_int_field(f, "timestamp", mf.timestamp); */
    put_hex_field(f, "certbody", mf.certbody, sizeof(mf.certbody));
    put_hex_field(f, "signature", mf.signature, sizeof(mf.signature));
    if(mf.signed_by != -1 || zeropoint_upgrade) {
/*      put_hex_field(f, "zp_pub", mf.zp_pub, sizeof(mf.zp_pub)); */
        put_hex_field(f, "zp_certbody", mf.zp_certbody, sizeof(mf.zp_certbody));
        put_hex_field(f, "zp_signature", mf.zp_signature,
                                         sizeof(mf.zp_signature));
    }
    fclose(f);
    res = 0;

quit:
    crypto_wipe(&mf, sizeof(mf));
    umask(saved_umask);
    dispose_fname(config_fname);
    dispose_fname(secretkey_fname);
    return res;
}


/* secret key isn't const because it gets wiped once used */
static int sign_and_save_point(const char *output_fn, const char *args_dir,
    int point, u_chr *secret_key, const u_chr *public_key, long timestamp)
{
    char *fname;
    FILE *f;
    struct config_file conf;
    unsigned char signer_key[secret_key_size];
    char output_fname[32]; /* 20 for node id, 7..9 for '_pNNN.key', 2 resvd */
    const char *ofn;
    int res;
    struct feda_cert_info certinfo;
    unsigned char signature[signature_size];
    int saved_umask;

    fname = make_config_fname(args_dir);
    res = read_config_file(fname, &conf);
    dispose_fname(fname);
    if(!res) {
        /* error message already printed by the function */
        message(mlv_info, "do you have the workplace deployed here?\n");
        return 1;
    }
    if(conf.point == -1) {
        message(mlv_alert, "can't zcrpoint/zsign at the master workplace\n");
        return 1;
    }
    if(conf.point != 0 && (
        all_zeroes(conf.zp_certbody, sizeof(conf.zp_certbody)) ||
        all_zeroes(conf.zp_signature, sizeof(conf.zp_signature))))
    {
        message(mlv_alert, "seems we don't have zero point deployed here\n");
        return 1;
    }

    compose_feda_cert(&certinfo,
                      conf.node_id, point, 0, timestamp, public_key);

    fname = conf.point == 0 ?
        make_secretkey_fname(args_dir) : make_zeropointkey_fname(args_dir);
    res = read_single_blob(fname, "secret", signer_key, sizeof(signer_key));
    dispose_fname(fname);
    if(!res) {
        message(mlv_alert, "can't read the file or there's no secret key\n");
        message(mlv_info, "can't get the zero point secret key\n");
        return 1;
    }

    crypto_eddsa_sign(signature, signer_key,
                      (unsigned char *)&certinfo, sizeof(certinfo));
    crypto_wipe(signer_key, sizeof(signer_key));


    /* okay, now try to save */

    if(output_fn) {
        ofn = output_fn;
    } else {
        hexdata2str(output_fname, conf.node_id, sizeof(conf.node_id));
        sprintf(output_fname+2*node_id_size, "_p%d.key", (int)point);
        ofn = output_fname;
    }

    saved_umask = umask(0077);
    f = fopen(ofn, "w");
    umask(saved_umask);
    if(!f) {
        message_perror(mlv_alert, "FATAL", ofn);
        return 1;
    }
 
    put_hex_field(f, "node_id", conf.node_id, node_id_size);
    put_int_field(f, "point", point);
    put_int_field(f, "minpref", conf.minpref);
    if(secret_key) {
        put_hex_field(f, "secret", secret_key, secret_key_size);
        crypto_wipe(secret_key, secret_key_size);
    }
    put_hex_field(f, "public", public_key, public_key_size);
    put_int_field(f, "signed_by", 0);
    put_hex_field(f, "master_pub", conf.master_pub, public_key_size);
    put_hex_field(f, "master_hash", conf.master_hash, yespower_hash_size);
    put_int_field(f, "timestamp", timestamp);
    put_hex_field(f, "certbody", &certinfo, sizeof(certinfo));
    put_hex_field(f, "signature", signature, sizeof(signature));
    if(conf.point == 0) {
        put_hex_field(f, "zp_pub", conf.public_key, public_key_size);
        put_hex_field(f, "zp_certbody", conf.certbody, feda_cert_size);
        put_hex_field(f, "zp_signature", conf.signature, signature_size);
    } else {
        put_hex_field(f, "zp_pub", get_cert_pubkey(conf.zp_certbody),
                                   public_key_size);
        put_hex_field(f, "zp_certbody", conf.zp_certbody, feda_cert_size);
        put_hex_field(f, "zp_signature", conf.zp_signature, signature_size);
    }
    fclose(f);

    return 0;
}

int main_zcrpoint(const struct fedakeys_args *args)
{
    int res;
    unsigned char seed[seed_size];
    unsigned char secret_key[secret_key_size];
    unsigned char public_key[public_key_size];
    long point;
    char *err;

    if(args->cmdargc < 2 || !args->cmdargs[1] || !*args->cmdargs[1]) {
        message(mlv_alert, "\"fedakeys zcrpoint\" needs the point id\n");
        return 1;
    }
    point = strtol(args->cmdargs[1], &err, 10);
    if(*err != 0 || point < 0 || point > 253) {
        message(mlv_alert, "invalid point id [%s]\n", args->cmdargs[1]);
        return 1;
    }

    res = get_random(seed, sizeof(seed));
    if(!res) {
        message(mlv_alert, "can't read random numbers\n");
        return 1;
    }

    crypto_eddsa_key_pair(secret_key, public_key, seed);

    return sign_and_save_point(args->out_file, args->dir,
                               point, secret_key, public_key, args->timestamp);
}

int main_zsign(const struct fedakeys_args *args)
{
    int res;
    unsigned char public_key[public_key_size];
    long point;
    char *err;

    if(args->cmdargc != 3 || !args->cmdargs[1] || !*args->cmdargs[1]) {
        message(mlv_alert, "\"fedakeys zsign\" needs exactly two args\n");
        return 1;
    }
    point = strtol(args->cmdargs[2], &err, 10);
    if(*err != 0 || point < 0 || point > 253) {
        message(mlv_alert, "invalid point id [%s]\n", args->cmdargs[2]);
        return 1;
    }

    res = read_single_blob(args->cmdargs[1], "public",
                           public_key, sizeof(public_key));
    if(!res) {
        message(mlv_alert, "can't read the file or there's no public key\n");
        message(mlv_info, "can't continue without the public key\n");
        return 1;
    }

    return sign_and_save_point(args->out_file, args->dir,
                               point, NULL, public_key, args->timestamp);
}

static void get_and_save_keypair(u_chr *seed, FILE *priv, FILE *pub)
{
    unsigned char secret_key[secret_key_size];
    unsigned char public_key[public_key_size];
    crypto_eddsa_key_pair(secret_key, public_key, seed);
    put_hex_field(priv, "secret", secret_key, sizeof(secret_key));
    crypto_wipe(secret_key, sizeof(secret_key));
    put_hex_field(pub, "public", public_key, sizeof(public_key));
}

int main_keypair(const struct fedakeys_args *args)
{
    unsigned char seed[seed_size];
    int res, saved_umask;

    if(args->cmdargc < 2 && !args->out_file) {
        message(mlv_alert, "\"fedakeys keypair\" needs a basename, try -h\n");
        return 1;
    }

    res = get_random(seed, sizeof(seed));
    if(!res) {
        message(mlv_alert, "can't read random numbers\n");
        return 1;
    }

    saved_umask = umask(077);

    if(args->out_file) {
        FILE *f;
        f = fopen(args->out_file, "w");
        if(!f) {
            message_perror(mlv_alert, "FATAL", args->out_file);
            res = 1;
            goto quit;
        }
        get_and_save_keypair(seed, f, f);
        fclose(f);
        res = 0;
    } else {
        FILE *fs, *fp;
        char *fname;
        int bnlen = strlen(args->cmdargs[1]);
        fname = malloc(bnlen + 5);
        strcpy(fname, args->cmdargs[1]);
        strcpy(fname + bnlen, ".key");
        fs = fopen(fname, "w");
        if(!fs) {
            message_perror(mlv_alert, "FATAL", fname);
            free(fname);
            res = 1;
            goto quit;
        }
        strcpy(fname + bnlen, ".pub");
        fp = fopen(fname, "w");
        if(!fs) {
            message_perror(mlv_alert, "FATAL", fname);
            free(fname);
            fclose(fs);
            res = 1;
            goto quit;
        }
        get_and_save_keypair(seed, fs, fp);
        fclose(fs);
        fclose(fp);
    }
    res = 0;
quit:
    umask(saved_umask);
    return res;
}

